from .Objects import *
from ..Sprite import Sprite
import pygame as pg
from .projection import Projection

class Space3D(Sprite):
    def __init__(self, par_surf, /,
                 real_width=WIDTH,
                 real_height=HEIGHT, **kwargs):
        super().__init__(par_surf, **kwargs)

        self.real_width, self.real_height = int(real_width), int(real_height)
        self.h_width, self.h_height = self.width // 2, self.height // 2

        self.workspace = (np.zeros((WIDTH, HEIGHT, 3), dtype=np.uint32),  # color
                          np.zeros((WIDTH, HEIGHT), dtype=float_bit),  # depth
                          np.zeros((WIDTH, HEIGHT), dtype=float_bit))  # priority

        self.last_drag_pos = (0, 0)
        self.last_mouse_wheel = 0
        self.enable = True

        self.mouse_sense = 0.1
        self.old_pressed_keys = []
        self.min_dist = 3

        self.camera = Camera(self, [0, 0, -10])
        self.projection = Projection(self)

        self.all_obj = []

    def load_object_from_file(self, filename, pos=None):
        vertex, faces = [], []
        with open(filename) as f:
            for line in f:
                if line.startswith('v '):
                    vertex.append([float(i) for i in line.split()[1:]] + [1])
                elif line.startswith('f'):
                    faces_ = line.split()[1:]
                    faces.append([int(face_.split('/')[0]) - 1 for face_ in faces_])
        self.all_obj.append(Solid3D(self, vertex, faces, pos))

    def clear_workspace(self):
        self.workspace = (np.zeros((WIDTH, HEIGHT, 3), dtype=np.uint32),  # color
                          np.zeros((WIDTH, HEIGHT), dtype=float_bit),  # depth
                          np.zeros((WIDTH, HEIGHT), dtype=float_bit))  # priority

    def draw(self):
        if self.enable:
            self.clear_workspace()
            for i in self.all_obj:
                i.vert_to_global()
            self.all_obj = sorted(self.all_obj, key=lambda x:np.linalg.norm(x.global_center_of_mass), reverse=True)
            for i in self.all_obj:
                i.draw(self.workspace)
            # maybe blit_array
            self.rect = self.surface.blit(
                pg.transform.scale(pg.surfarray.make_surface(self.workspace[0]), (self.width, self.height)), self.pos)

    def add_object(self, obj):
        self.all_obj.append(obj)

    def update(self):
        if self.rect.collidepoint(pg.mouse.get_pos()):
            self.camera.control()
            self.update_keys()

        delta = self.last_mouse_wheel - self.par_surf.mouse_wheel_pos
        self.last_mouse_wheel = self.par_surf.mouse_wheel_pos
        if self.camera.mode == 1 and delta:
            self.camera.control(transition=np.array([0, 0, -delta, 0]))

    def dragged(self, *args, **kwargs):
        if kwargs["btn_id"] == 1:
            rotation = kwargs["mouse_delta"][0] * self.mouse_sense, kwargs["mouse_delta"][1] * self.mouse_sense, 0
        else:
            rotation = 0, 0, kwargs["mouse_delta"][0] * self.mouse_sense

        self.camera.control(rotation=rotation)

    def slip(self, *args, **kwargs):
        self.camera.control(transition=np.array([0, 0, 0, 0]), rotation=[0, 0, 0])

    def update_keys(self):
        pressed_keys = self.par_surf.pressed_keys[:]
        transition = np.array([0, 0, 0, 0])
        if pg.K_a in pressed_keys:
            transition[0] = -1
        if pg.K_d in pressed_keys:
            transition[0] = 1
        if pg.K_w in pressed_keys:
            transition[2] = 1
        if pg.K_s in pressed_keys:
            transition[2] = -1
        if pg.K_LSHIFT in pressed_keys:
            transition[1] = 1
        if pg.K_LCTRL in pressed_keys:
            transition[1] = -1
        if self.old_pressed_keys != pressed_keys:
            if self.camera.mode == 0:
                self.camera.control(transition=transition)
            self.old_pressed_keys = pressed_keys[:]



import pygame as pg

from .Sprite import Sprite


class Text(Sprite):
    def __init__(self, par_surf, /,
                 inp_text=lambda *args: "your text",
                 color=(255, 255, 255),
                 font='serif',
                 font_size=10, **kwargs):
        super().__init__(par_surf, color=color, **kwargs)
        self.inp_text = inp_text
        self.text = pg.font.SysFont(font, int(font_size * self.ps_height / 500))
        self.txt_render = self.text.render(str(self.inp_text()), True, self.color)

    def pressed(self, *args, **kwargs):
        self.func(args, **kwargs)

    def update(self):
        self.txt_render = self.text.render(str(self.inp_text()), True, self.color)

    def draw(self):
        self.rect = self.surface.blit(self.txt_render, self.pos)



import numpy as np
from numba import njit

from .Objects.pre_settings import *


@njit(fastmath=True)
def triangle_area(ax, ay, bx, by, cx, cy):
    # return (np.cross(b - a, c - a)) / 2
    abx, aby = bx - ax, by - ay
    acx, acy = cx - ax, cy - ay
    return (abx * acy - aby * acx) / 2


@njit(fastmath=True)
def range_cut(mi, ma, num):
    return min(max(num, mi), ma)


@njit(fastmath=True)
def set_pixel(color_mat, depth_mat, priority_mat, x, y, color, depth, priority):
    color_mat[x, y, :] = color
    depth_mat[x, y] = depth
    priority_mat[x, y] = priority


@njit(fastmath=True)
def render_vertices(color_mat, depth_mat, priority_mat, vertices, vertex_colors, vertex_radius):
    for vertex in range(len(vertices)):
        x0, y0, z0, w0 = vertices[vertex]
        if w0 <= 0:
            continue
        x0, y0, z0 = x0 / w0, y0 / w0, w0
        xp0 = int((x0 + 1) / 2 * WIDTH)
        yp0 = int((1 - y0) / 2 * HEIGHT)
        if not (0 <= xp0 < WIDTH and 0 <= yp0 < HEIGHT):
            continue
        rad = vertex_radius[vertex]
        dist = z0

        for dx in range(-rad, rad):
            for dy in range(-rad, rad):
                cx, cy = xp0 + dx, yp0 + dy
                if 0 <= cx < WIDTH and 0 <= cy < HEIGHT and dx ** 2 + dy ** 2 < rad ** 2:
                    if depth_mat[cx, cy] + jitter > dist > depth_mat[cx, cy] - jitter:
                        if priority_mat[cx, cy] < dist and priority_mat[cx, cy] != 0:
                            continue
                    elif depth_mat[cx, cy] < dist and depth_mat[cx, cy] != 0:
                        continue
                    set_pixel(color_mat, depth_mat, priority_mat, cx, cy, vertex_colors[vertex], z0, z0)
                    # color_mat[cx, cy, :] = vertex_colors[vertex]
                    # depth_mat[cx, cy] = z0
                    # priority_mat[cx, cy] = z0


@njit(fastmath=True)
def render_edges(color_mat, depth_mat, priority_mat, vertices, edges, edge_colors, thickness):
    for edge in range(len(edges)):
        all_vert = vertices[edges[edge]]
        th = thickness[edge]
        if len(all_vert) < 2:
            continue
        x0, y0, z0, w0 = all_vert[0]
        x1, y1, z1, w1 = all_vert[1]
        w0 = float(w0)
        w1 = float(w1)
        if w0 <= 1e-5 or w1 <= 1e-5 or w0 <= 1e-5 or w1 <= 1e-5:
            continue
        x0, y0, z0 = x0 / w0, y0 / w0, w0
        x1, y1, z1 = x1 / w1, y1 / w1, w1
        xl, yl, zl = x1 - x0, y1 - y0, z1 - z0
        if 1e-8 > yl > -1e-8 and 1e-8 > xl > -1e-8:
            continue
        xp0 = int((x0 + 1) / 2 * WIDTH)
        xp1 = int((x1 + 1) / 2 * WIDTH)
        yp0 = int((1 - y0) / 2 * HEIGHT)
        yp1 = int((1 - y1) / 2 * HEIGHT)
        bbx_min = int(max(0, min(xp0, xp1)))
        bbx_max = int(min(WIDTH - 1, max(xp0, xp1)))
        bby_min = int(max(0, min(yp0, yp1)))
        bby_max = int(min(HEIGHT - 1, max(yp0, yp1)))
        if bby_max - bby_min > bbx_max - bbx_min:
            if 1e-8 > yl > -1e-8:
                continue
            for cy in range(bby_min, bby_max):
                yp = 1 - cy / HEIGHT * 2
                xp = xl * (yp - y0) / yl + x0
                if not 1 > xp > -1:
                    continue
                cx = int((xp + 1) / 2 * WIDTH)
                dist = zl * (yp - y0) / yl + z0
                if depth_mat[cx, cy] + jitter > dist > depth_mat[cx, cy] - jitter:
                    if priority_mat[cx, cy] < dist and priority_mat[cx, cy] != 0:
                        continue
                elif depth_mat[cx, cy] < dist and depth_mat[cx, cy] != 0:
                    continue
                for t in range(-th // 2, th // 2):
                    ind = range_cut(0, WIDTH - 1, int(cx + t))
                    set_pixel(color_mat, depth_mat, priority_mat, ind, cy, edge_colors[edge], dist, dist)
                    # color_mat[ind, cy, :] = edge_colors[edge]
                    # color_mat[ind, cy, :] = [zp * 50, zp * 50, zp * 50]
                    # depth_mat[ind, cy] = zp
                    # priority_mat[cx, cy] = zp
        else:
            if 1e-8 > xl > -1e-8:
                continue
            for cx in range(bbx_min, bbx_max):
                xp = cx / WIDTH * 2 - 1
                yp = yl * (xp - x0) / xl + y0
                if not 1 > yp > -1:
                    continue
                cy = int((1 - yp) / 2 * HEIGHT)
                dist = zl * (xp - x0) / xl + z0
                if depth_mat[cx, cy] + jitter > dist > depth_mat[cx, cy] - jitter:
                    if priority_mat[cx, cy] < dist and priority_mat[cx, cy] != 0:
                        continue
                elif depth_mat[cx, cy] < dist and depth_mat[cx, cy] != 0:
                    continue
                for t in range(-th // 2, th // 2):
                    ind = range_cut(0, HEIGHT - 1, int(cy + t))
                    set_pixel(color_mat, depth_mat, priority_mat, cx, ind, edge_colors[edge], dist, dist)
                    # color_mat[cx, ind, :] = edge_colors[edge]
                    # color_mat[cx, ind, :] = [zp*50, zp*50, zp*50]

                    # depth_mat[cx, ind] = zp
                    # priority_mat[cx, cy] = z0


@njit(fastmath=True)
def render_polygons(color_mat, depth_mat, priority_mat, vertices, faces, normals, face_colors):
    for face in range(len(faces)):
        all_vert = vertices[faces[face]]
        x0, y0, z0, w0 = all_vert[0]
        x1, y1, z1, w1 = all_vert[1]
        x2, y2, z2, w2 = all_vert[2]
        neg_points = np.array([[0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]).astype(float_bit)
        positive_points = np.array([[0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]).astype(float_bit)
        negative_counter = 0
        positive_counter = 0
        vert = np.zeros((3, 3)).astype(float_bit)
        if w0 <= NEAR_PLANE:
            neg_points[negative_counter] = all_vert[0]
            negative_counter += 1
        else:
            positive_points[positive_counter] = all_vert[0]
            positive_counter += 1

        if w1 <= NEAR_PLANE:
            neg_points[negative_counter] = all_vert[1]
            negative_counter += 1
        else:
            positive_points[positive_counter] = all_vert[1]
            positive_counter += 1

        if w2 <= NEAR_PLANE:
            neg_points[negative_counter] = all_vert[2]
            negative_counter += 1
        else:
            positive_points[positive_counter] = all_vert[2]
            positive_counter += 1

        if negative_counter == 3:
            continue
        elif negative_counter == 2:
            x0, y0, z0, w0 = positive_points[0]
            x1, y1, z1, w1 = neg_points[0]
            x2, y2, z2, w2 = neg_points[1]
            z0, z1, z2 = w0, w1, w2

            x1, y1 = (NEAR_PLANE - z0) / (z1 - z0) * (x1 - x0) + x0, (NEAR_PLANE - z0) / (z1 - z0) * (y1 - y0) + y0
            x2, y2 = (NEAR_PLANE - z0) / (z2 - z0) * (x2 - x0) + x0, (NEAR_PLANE - z0) / (z2 - z0) * (y2 - y0) + y0
            z1 = z2 = NEAR_PLANE
            # x0, y0 = x0 / z0, y0 / z0
            # x1, y1 = x1 / z1, y1 / z1
            # x2, y2 = x2 / z2, y2 / z2
            vert[0, 0], vert[0, 1], vert[0, 2], vert[1, 0], vert[1, 1], vert[1, 2], vert[2, 0], vert[2, 1], vert[
                2, 2] = x0, y0, z0, x1, y1, z1, x2, y2, z2
            draw_face(color_mat, depth_mat, priority_mat, vert, normals[face], face_colors[face])

        elif negative_counter == 1:
            x0, y0, z0, w0 = neg_points[0]
            x1, y1, z1, w1 = positive_points[0]
            x2, y2, z2, w2 = positive_points[1]
            z0, z1, z2 = w0, w1, w2

            x3, y3 = (NEAR_PLANE - z0) / (z1 - z0) * (x1 - x0) + x0, (NEAR_PLANE - z0) / (z1 - z0) * (y1 - y0) + y0
            x4, y4 = (NEAR_PLANE - z0) / (z2 - z0) * (x2 - x0) + x0, (NEAR_PLANE - z0) / (z2 - z0) * (y2 - y0) + y0
            z3 = z4 = NEAR_PLANE
            # x1, y1 = x1 / z1, y1 / z1
            # x2, y2 = x2 / z2, y2 / z2
            # x3, y3 = x3 / z3, y3 / z3
            # x4, y4 = x4 / z4, y4 / z4

            vert[0, 0], vert[0, 1], vert[0, 2], vert[1, 0], vert[1, 1], vert[1, 2], vert[2, 0], vert[2, 1], vert[
                2, 2] = x1, y1, z1, x2, y2, z2, x3, y3, z3
            draw_face(color_mat, depth_mat, priority_mat, vert, normals[face], face_colors[face])
            vert[0, 0], vert[0, 1], vert[0, 2], vert[1, 0], vert[1, 1], vert[1, 2], vert[2, 0], vert[2, 1], vert[
                2, 2] = x2, y2, z2, x3, y3, z3, x4, y4, z4
            draw_face(color_mat, depth_mat, priority_mat, vert, normals[face], face_colors[face])
        elif negative_counter == 0:
            x0, y0, z0, w0 = positive_points[0]
            x1, y1, z1, w1 = positive_points[1]
            x2, y2, z2, w2 = positive_points[2]
            z0, z1, z2 = w0, w1, w2
            # x0, y0 = x0 / z0, y0 / z0
            # x1, y1 = x1 / z1, y1 / z1
            # x2, y2 = x2 / z2, y2 / z2

            vert[0, 0], vert[0, 1], vert[0, 2], vert[1, 0], vert[1, 1], vert[1, 2], vert[2, 0], vert[2, 1], vert[
                2, 2] = x0, y0, z0, x1, y1, z1, x2, y2, z2
            draw_face(color_mat, depth_mat, priority_mat, vert, normals[face], face_colors[face])


@njit(fastmath=True)
def draw_face(color_mat, depth_mat, priority_mat, vertices, g_normal, face_color):
    x0, y0, z0 = vertices[0]
    x1, y1, z1 = vertices[1]
    x2, y2, z2 = vertices[2]
    cent = (z0 + z1 + z2) / 3
    x0, y0 = x0 / z0, y0 / z0
    x1, y1 = x1 / z1, y1 / z1
    x2, y2 = x2 / z2, y2 / z2
    plane_a = (y0 - y1) * (z0 - z2) - (z0 - z1) * (y0 - y2)
    plane_b = -(x0 - x1) * (z0 - z2) + (z0 - z1) * (x0 - x2)
    plane_c = (x0 - x1) * (y0 - y2) - (y0 - y1) * (x0 - x2)
    plane_d = plane_a * x0 + plane_b * y0 + plane_c * z0
    xp0 = int((x0 + 1) / 2 * WIDTH)
    xp1 = int((x1 + 1) / 2 * WIDTH)
    xp2 = int((x2 + 1) / 2 * WIDTH)
    yp0 = int((1 - y0) / 2 * HEIGHT)
    yp1 = int((1 - y1) / 2 * HEIGHT)
    yp2 = int((1 - y2) / 2 * HEIGHT)

    if plane_c == 0:
        return
    bbx_min = int(max(0, min(min(xp0, xp1), xp2)))
    bbx_max = int(min(WIDTH - 1, max(max(xp0, xp1), xp2)))
    bby_min = int(max(0, min(min(yp0, yp1), yp2)))
    bby_max = int(min(HEIGHT - 1, max(max(yp0, yp1), yp2)))
    lighting = (np.dot(g_normal, LIGHT_DIRECTION) / (
            np.linalg.norm(LIGHT_DIRECTION, ord=1) * np.linalg.norm(g_normal, ord=1)) + 1) / 2
    r, g, b = face_color
    h, s, l = rgb_to_hsl(r, g, b)
    color = np.array(hsl_to_rgb(h, s, lighting ** 2))

    for cx in range(bbx_min, bbx_max):
        for cy in range(bby_min, bby_max):
            xp = float(cx) * 2 / WIDTH - 1
            yp = 1 - float(cy) * 2 / HEIGHT
            dist = -(plane_a * xp + plane_b * yp - plane_d) / plane_c
            if depth_mat[cx, cy] + jitter > dist > depth_mat[cx, cy] - jitter:
                if priority_mat[cx, cy] < cent and priority_mat[cx, cy] != 0:
                    continue
            elif depth_mat[cx, cy] < dist and depth_mat[cx, cy] != 0:
                continue
            full_triangle_area = abs(triangle_area(xp0, yp0, xp1, yp1, xp2, yp2))
            area_sum = 0
            area_sum += abs(triangle_area(xp0, yp0, xp1, yp1, cx, cy))
            area_sum += abs(triangle_area(xp0, yp0, xp2, yp2, cx, cy))
            area_sum += abs(triangle_area(xp1, yp1, xp2, yp2, cx, cy))
            if area_sum + 0.01 > full_triangle_area > area_sum - 0.01:
                set_pixel(color_mat, depth_mat, priority_mat, cx, cy, color, dist, cent)


@njit(fastmath=True)
def rgb_to_hsl(r, g, b):
    high = max(r, g, b)
    low = min(r, g, b)
    l = ((high + low) / 2)

    if high == low:
        h = 0.0
        s = 0.0
    else:
        d = high - low
        s = d / (2 - high - low) if l > 0.5 else d / (high + low)
        h = {
            r: (g - b) / d + (6 if g < b else 0),
            g: (b - r) / d + 2,
            b: (r - g) / d + 4,
        }[high]
        h /= 6
    l /= 255

    return h, s, l


@njit(fastmath=True)
def hsl_to_rgb(h, s, l):
    def hue_to_rgb(p, q, t):
        t += 1 if t < 0 else 0
        t -= 1 if t > 1 else 0
        if t < 1 / 6: return p + (q - p) * 6 * t
        if t < 1 / 2: return q
        if t < 2 / 3: p + (q - p) * (2 / 3 - t) * 6
        return p

    if s == 0:
        r, g, b = l, l, l
    else:
        q = l * (1 + s) if l < 0.5 else l + s - l * s
        p = 2 * l - q
        r = hue_to_rgb(p, q, h + 1 / 3)
        g = hue_to_rgb(p, q, h)
        b = hue_to_rgb(p, q, h - 1 / 3)

    return int(r * 255), int(g * 255), int(b * 255)


@njit(fastmath=True)
def overlap(a, b):
    for i in a:
        for j in b:
            if i == j:
                return True
    return False


@njit(fastmath=True)
def face_fast_sort(ref, points):
    return np.argsort(np.array(list([np.linalg.norm(ref - points[i][:3]) for i in range(len(points))])))


@njit(fastmath=True)
def detect_not_drawn_vertices(vertices, md, not_drawn_vertices):
    l = 0
    for v in range(len(vertices)):
        if vertices[v][-1] < 0:
            not_drawn_vertices[l] = v
            l += 1
            continue
        else:
            vertices[v] /= vertices[v][-1]
            x, y, z = vertices[v][:3]
            if not (md > x > -md and md > y > -md and md > z > -md):
                not_drawn_vertices[l] = v
                l += 1
    return l



from .. fast_math import *
from .. matrix_functions import *

class Object3D:
    def __init__(self, render, /,
                 vertices=None, vertex_radius=None, vertex_colors=None,
                 edges=None, edges_thickness=None, color_edges=None,
                 faces=None, color_faces=None,
                 pos=None):

        if not pos:
            pos = [0, 0, 0]
        self.render = render

        x, y, z = pos
        self.transform = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [x, y, z, 1]
        ]).astype(float_bit)
        self.post_projection_transform = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1]
        ]).astype(float_bit)

        self.global_vert = np.array(False)
        self.global_center_of_mass = np.array(False)

        self.draw_vertices = True
        self.draw_edges = True
        self.draw_faces = True
        self.fix_to_camera = False
        self.vertices_on_screen = np.array(False)

        self.empty = True

        if not isinstance(vertices, np.ndarray):
            vertices = np.array([np.array(v) for v in vertices]).astype(float_bit)
        if vertices.any():
            self.vertices = np.array([np.array(v).astype(float_bit) for v in vertices]).astype(float_bit)
            if vertex_colors:
                self.vertex_colors = np.array(vertex_colors).astype(np.int16)
            else:
                self.vertex_colors = np.array([[255, 255, 255] for _ in self.vertices]).astype(np.int32)
            if vertex_radius:
                self.vertex_radius = np.array(vertex_radius).astype(np.int16)
            else:
                self.vertex_radius = np.array([10] * len(self.vertices)).astype(np.int16)

            self.global_vert = np.copy(self.vertices)
            self.center_of_mass = np.mean(self.vertices, axis=0).astype(float_bit)
            self.empty = False
        else:
            self.draw_vertices = False
            self.draw_edges = False
            self.draw_faces = False

        if edges:
            self.edges = np.array(edges).astype(np.uint32)
            if edges_thickness:
                self.edges_thickness = np.array(edges_thickness).astype(np.int16)
            else:
                self.edges_thickness = np.array([1] * len(self.edges)).astype(np.int16)

            if color_edges:
                self.color_edges = np.array(color_edges).astype(np.int32)
            else:
                self.color_edges = np.array([[255, 255, 255] for _ in self.edges]).astype(np.int32)
        else:
            self.draw_edges = False

        if faces:
            faces_ = []
            for face in faces:
                for f_ in range(2, len(face)):
                    faces_.append([face[f_ - 1], face[f_], face[0]])
            self.faces = np.array(faces_).astype(np.uint32)
            if color_faces:
                self.color_faces = np.array(color_faces).astype(np.uint32)
            else:
                self.color_faces = np.array([[255, 255, 255] for _ in self.faces]).astype(np.uint32)
            face_normals = []
            for face in self.faces:
                A = self.vertices[face[0]][:3] - self.vertices[face[-1]][:3]
                B = self.vertices[face[0]][:3] - self.vertices[face[1]][:3]
                face_normals.append(np.cross(B, A))
            self.face_normals = np.array(face_normals).astype(float_bit)
        else:
            self.draw_faces = False

        self.render.add_object(self)
    def translate(self, pos):
        self.transform = self.transform @ translate(pos)

    def scale(self, n1, n2, n3):
        self.transform = self.transform @ scale(n1, n2, n3)

    def rotate_x(self, angle):
        self.transform = self.transform @ rotate_x(angle)

    def rotate_y(self, angle):
        self.transform = self.transform @ rotate_y(angle)

    def rotate_z(self, angle):
        self.transform = self.transform @ rotate_z(angle)

    def position(self):
        return self.transform[:3, 3]

    def vert_to_global(self):
        if not self.empty:
            self.global_vert = self.vertices @ self.transform
            self.global_center_of_mass = self.center_of_mass @ self.transform

    def draw(self, workspace):
        if self.empty:
            return
        if not self.fix_to_camera:
            vertices = self.global_vert @ self.render.camera.camera_matrix()
        else:
            vertices = self.global_vert

        vertices = vertices @ self.render.projection.projection_matrix
        vertices = vertices @ self.post_projection_transform
        if self.draw_vertices:
            render_vertices(*workspace,
                            vertices,
                            self.vertex_colors,
                            self.vertex_radius)
        if self.draw_edges:
            self.vertices_on_screen = vertices[:, :2]/np.array([vertices[:, 2], vertices[:, 2]]).T
            render_edges(*workspace,
                         vertices,
                         self.edges,
                         self.color_edges,
                         self.edges_thickness
                         )
        if self.draw_faces:
            render_polygons(*workspace,
                            vertices,
                            self.faces,
                            self.face_normals,
                            self.color_faces
                            )




import numpy as np
import pygame as pg
from ..matrix_functions import *
from .pre_settings import *


class Camera:
    def __init__(self, render, position):
        self.render = render
        self.init_pos = position

        self.transition = np.array([0, 0, 0, 0])
        self.rotation = [0, 0]
        self.h_fov = math.pi / 3
        self.v_fov = self.h_fov * (render.height / render.width)
        self.near_plane = NEAR_PLANE
        self.far_plane = FAR_PLANE
        self.moving_speed = 0.9
        self.zoom_speed = 3
        self.rotation_speed = 0.011
        self.mode = 1
        self.direction = np.eye(4)
        self.roll = 0
        self.pitch = 0
        self.yaw = 0

        # service variables
        self.pos_serv = np.array([*position, 1.0])
        self.pos_serv_1 = np.array([*position, 1.0])
        self.camera_matrix_0_serv = np.eye(4)
        self.camera_matrix_1_serv = np.eye(4)

    def reset(self):
        self.pos_serv = np.array([*self.init_pos, 1.0])
        self.transition = np.array([0, 0, 0, 0])
        self.rotation = [0, 0]
        self.near_plane = NEAR_PLANE
        self.far_plane = FAR_PLANE
        self.moving_speed = 0.9
        self.zoom_speed = 3
        self.rotation_speed = 0.011
        self.direction = np.eye(4)

    def control(self, /, transition=None, rotation=None):
        if isinstance(transition, np.ndarray):
            self.transition = transition * self.moving_speed

        if rotation:
            self.rotation = rotation
        else:
            self.rotation = [0, 0, 0]

        if self.mode == 0:
            self.pos_serv += self.transition @ self.direction * self.moving_speed
        elif self.mode == 1 and isinstance(transition, np.ndarray):
            self.pos_serv += self.transition * self.zoom_speed
        if self.mode == 0:
            if self.rotation[0]:
                self.pitch += self.rotation[0] * self.rotation_speed
            if self.rotation[1]:
                self.roll += self.rotation[1] * self.rotation_speed
            #if self.rotation[2]:
            #    self.yaw += self.rotation[2] * self.rotation_speed
            self.direction = rotate_x(self.roll) @ rotate_y(self.pitch) @ rotate_z(self.yaw)
        elif self.mode == 1:
            if self.rotation[0]:
                self.direction = (rotate_y(self.rotation[0] * self.rotation_speed * -1)) @ self.direction
            if self.rotation[1]:
                self.direction = (rotate_x(self.rotation[1] * self.rotation_speed * -1)) @ self.direction
            if self.rotation[2]:
                self.direction = (rotate_z(self.rotation[2] * self.rotation_speed * -1)) @ self.direction



        # make matrix
        x, y, z, w = self.pos_serv
        translate_matrix = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [-x, -y, -z, 1]
        ])
        dir_t = self.direction.T
        if self.mode == 0:
            self.camera_matrix_0_serv = translate_matrix @ dir_t
        elif self.mode == 1:
            self.camera_matrix_1_serv = dir_t @ translate_matrix
            self.pos_serv_1 = dir_t @ self.pos_serv

    def camera_matrix(self):
        if self.mode == 0:
            return self.camera_matrix_0_serv
        elif self.mode == 1:
            return self.camera_matrix_1_serv

    def position(self):
        if self.mode == 0:
            return self.pos_serv
        elif self.mode == 1:
            return self.pos_serv_1



import pygame as pg

class NoCvMatSet(Exception):
    """Raised when no cv mat set for object of class Mat"""
    pass


class NoDrawFuncReloaded(Exception):
    """Raised when sprite draw func is not reloaded in child class"""
    pass


class Sprite:
    def __init__(self, par_surf, /,
                 name="Sprite",
                 x=0,
                 y=0,
                 width=0,
                 height=0,
                 func=lambda *args, **kwargs: args,
                 color=(100, 100, 100),
                 transparent_for_mouse=False,
                 **kwargs):
        self.func = func
        self.name = name
        self.par_surf = par_surf
        self.par_surf.sprite(self)
        self.surface = self.par_surf.screen
        self.ps_width, self.ps_height = par_surf.width, par_surf.height
        self.x = int(x * self.ps_width)
        self.y = int(y * self.ps_height)
        self.pos = self.x, self.y
        self.width = int(width * self.ps_width)
        self.height = int(height * self.ps_height)
        self.color = color
        self.rect = pg.Rect(self.x, self.y, self.width, self.height)
        self.transparent_for_mouse = transparent_for_mouse
        self.visible = True

    def pressed(self, *args, **kwargs):
        pass

    def dragged(self, *args, **kwargs):
        pass

    def release(self, *args, **kwargs):
        pass

    def hover(self, *args, **kwargs):
        pass

    def update(self):
        pass

    def convert_to_local(self, coords):
        return coords[0] - self.x, coords[1] - self.y

    def set_pos(self, pos):
        self.pos = pos[0]*self.ps_width, pos[1]*self.ps_height
        self.x = pos[0]*self.ps_width
        self.y = pos[1]*self.ps_height

    def draw(self):
        raise NoDrawFuncReloaded



from .Sprite import Sprite
import pygame as pg


class Button(Sprite):
    def __init__(self, par_surf, /, image="", radius=10, **kwargs):
        super().__init__(par_surf, **kwargs)
        self.radius = radius
        if image:

            if isinstance(image, str):
                self.have_image = 1
                self.image = pg.transform.scale(pg.image.load(image), (self.width, self.height))

            else:
                self.have_image = len(image)
                self.image_arr = []
                for img in image:
                    self.image_arr.append(pg.transform.scale(pg.image.load(img), (self.width, self.height)))
                self.image = self.image_arr[0]
                self.curr_img = 0
        else:
            self.have_image = 0

    def pressed(self, *args, **kwargs):
        self.func(*args, **kwargs)

    def draw(self):
        if self.have_image == 1:
            self.surface.blit(self.image, self.rect)
        elif self.have_image > 1:
            self.surface.blit(self.image_arr[self.curr_img], self.rect)
        else:
            pg.draw.rect(self.surface, self.color, self.rect, border_radius=self.radius)



__all__ = ["Slider"]

from .Sprite import Sprite
import pygame as pg


def range_cut(mi, ma, val):
    return min(ma, max(mi, val))


def convert_range(new_min, new_max, old_min, old_max, old_value):
    old_range = (old_max - old_min)
    new_range = (new_max - new_min)
    return (((old_value - old_min) * new_range) / old_range) + new_min


class Slider(Sprite):
    def __init__(self, par_surf, /,
                 slider_color=(255, 255, 255),
                 min=0,
                 max=100,
                 val=None, **kwargs):
        super().__init__(par_surf, **kwargs)
        self.slider_color = slider_color
        self.min = min
        self.max = max
        self.slider_rad = self.height // 2
        self.slider_y = self.slider_rad

        if val:
            self.val = val
        else:
            self.val = self.min

        self.slider_x = convert_range(self.slider_rad, self.width - self.slider_rad, self.min, self.max, self.val)

    def set_val(self, val):
        self.slider_x = convert_range(self.slider_rad, self.width - self.slider_rad, self.min, self.max, val)

    def dragged(self, *args, **kwargs):
        pos = kwargs["mouse_pos"]
        self.slider_x = range_cut(self.slider_rad, self.width - self.slider_rad, pos[0])
        self.val = convert_range(self.min, self.max, self.slider_rad, self.width - self.slider_rad, self.slider_x)
        self.func(self.val)

    def update(self):
        pg.draw.rect(self.surface, self.color, (0, 0, self.width, self.height), border_radius=self.height // 2)
        pg.draw.circle(self.surface, (255, 255, 255), (self.slider_x, self.slider_y), self.slider_rad)

    def draw(self):
        self.slider_x = convert_range(self.slider_rad, self.width - self.slider_rad, self.min, self.max, self.val)
        pg.draw.rect(self.surface, self.color, (self.x, self.y, self.width, self.height),
                     border_radius=self.height // 2)
        pg.draw.circle(self.surface, (255, 255, 255), (self.x + self.slider_x, self.y + self.slider_y), self.slider_rad)




from .Object3D import Object3D


class Solid3D(Object3D):
    def __init__(self, render, vertices='', faces='', **kwargs):
        super().__init__(render, vertices=vertices, faces=faces, **kwargs)
        self.draw_vertices = False
        self.draw_edges = False


from .Object3D import Object3D


class Dot3D(Object3D):
    def __init__(self, render, vertices='', edges='', **kwargs):
        super().__init__(render, vertices=vertices, edges=edges, **kwargs)
        self.draw_edges = False
        self.draw_faces = False



from .Sprite import Sprite
from .Sprite import NoCvMatSet
import pygame as pg

class Mat(Sprite):
    def __init__(self, par_surf, /,
                 cv_mat_stream=None,
                 **kwargs):

        if cv_mat_stream:
            self.cv_mat_stream = cv_mat_stream
        else:
            raise NoCvMatSet
        super().__init__(par_surf, **kwargs)
        self.is_mat_stream = False
        self.last_hover_pos = (0, 0)
        self.is_pressed = False

    def draw(self):
        mat = self.cv_mat_stream()
        if self.width != 0 and self.height != 0:
            self.rect = self.surface.blit(pg.transform.flip(
                pg.transform.scale(pg.transform.rotate(pg.surfarray.make_surface(mat), -90), (self.width, self.height)),
                1, 0), self.pos)
        else:
            self.rect = self.surface.blit(
                pg.transform.flip(pg.transform.rotate(pg.surfarray.make_surface(mat), -90), 1, 0),
                self.pos)

    def update(self):
        self.func(mouse_pos=self.last_hover_pos, btn_id=self.is_pressed)

    def pressed(self, *args, **kwargs):
        self.last_hover_pos = kwargs["mouse_pos"]
        self.is_pressed = kwargs["btn_id"]

    def dragged(self, *args, **kwargs):
        self.last_hover_pos = kwargs["mouse_pos"]
        self.is_pressed = kwargs["btn_id"]
        pass

    def hover(self, *args, **kwargs):
        self.last_hover_pos = kwargs["mouse_pos"]
        self.is_pressed = False



import numpy as np
from numba import njit

WIDTH = 300
HEIGHT = 300

jitter = 0.001


NEAR_PLANE = 0.1
FAR_PLANE = 100
float_bit = np.float64
LIGHT_DIRECTION = np.array([1, 0.6, 0.3]).astype(float_bit)




import pygame as pg
import threading as thr
from .Sprites import *


class Screen:
    def __init__(self, width, height, fps=60):
        self.objects = dict()
        self.objects_order = []
        self.raw_draw = dict()
        self.blit_arr = []
        self.width, self.height = width, height
        self.screen = None
        self.pressed_obj_ind = None
        self.fps = 0
        self.pressed_keys = []
        self.curr_obj = None
        self.clock = pg.time.Clock()
        self.max_fps = fps
        self.bg_color = (0,0,0)

        self.mouse_delta = [0, 0]
        self.process_btn_clk = 0
        self.pressed_obj = None
        self.released_obj = None
        self.dragged_obj = None
        self.old_mouse_pos = [0, 0]
        self.mouse_wheel_pos = 0
        self.mouse_pos = [0, 0]
        self.mouse_state = [0, 0]
        self.initialized = False
        self.in_separate_thread = False
        self.force_stop = False
        self.del_queue = []
        self.cicle_finished = True

    def lunch_separate_thread(self):
        self.in_separate_thread=True
        self.thread = thr.Thread(target=self.as_thread)
        self.thread.start()

    def as_thread(self):
        self.init()
        self.run()

    def __getitem__(self, item):
        if item in self.objects:
            return self.objects[item]
        else:
            raise AttributeError("No object named", item)

    def add_fps_indicator(self, x=0, y=0, font=10, color=(255,255,255)):
        Text(self, name="fps", x=x, y=y, inp_text=self.get_fps, font='serif', font_size=font, color=color)


    def init(self):
        pg.init()
        self.screen = pg.display.set_mode((self.width, self.height))
        self.initialized = True


    def step(self):
        if self.running:
            self.handle_events()
            self.update_mouse()
            self.update_objects()
            self.draw_objects()
            pg.display.update()
            pg.display.flip()
            self.clock.tick(self.max_fps)
            self.fps = self.clock.get_fps()
            self.post_processes()
            self.cicle_finished = True

    def run(self):
        while self.running:
            self.step()
        pg.quit()

    def end(self):
        self.running = False
        pg.quit()


    @property
    def running(self):
        if self.in_separate_thread:
            return thr.main_thread().is_alive()
        else:
            return not self.force_stop
    @running.setter
    def running(self, val):
        self.force_stop = val

    def wait_step(self):
        self.cicle_finished = False
        while not self.cicle_finished:
            pass

    def get_fps(self):
        return round(self.fps, 2)

    def update_mouse(self):
        self.mouse_pos = pg.mouse.get_pos()
        self.mouse_delta = [self.old_mouse_pos[0] - self.mouse_pos[0], self.old_mouse_pos[1] - self.mouse_pos[1]]
        self.old_mouse_pos = self.mouse_pos[:]


    def handle_events(self):
        for obj_name in self.objects_order[::-1]:
            obj = self.objects[obj_name]
            if obj.rect.collidepoint(self.mouse_pos) and not obj.transparent_for_mouse:
                self.curr_obj = obj
                break
            self.curr_obj = None

        for event in pg.event.get():
            if event.type == pg.QUIT:
                self.running = False
                return
            elif event.type == pg.KEYDOWN:
                if event.key == pg.K_ESCAPE:
                    self.running = False
                    return
                else:
                    self.pressed_keys.append(event.key)
            elif event.type == pg.KEYUP:
                if event.key in self.pressed_keys:
                    self.pressed_keys.pop(self.pressed_keys.index(event.key))
            elif event.type == pg.MOUSEWHEEL:
                self.mouse_wheel_pos += event.y
            elif event.type == pg.MOUSEBUTTONDOWN:
                self.update_mouse()
                self.mouse_state = [event.button, 1]
                self.pressed_obj = self.curr_obj
            elif event.type == pg.MOUSEBUTTONUP:
                self.mouse_state = [event.button, 0]
                self.released_obj = self.curr_obj


    def add_object(self, obj):
        if obj.name in self.objects:
            raise ValueError("Duplicate object")
        self.objects[obj.name] = obj
        self.objects_order.append(obj.name)

    def del_object(self, obj):
        if isinstance(obj, str):
            obj_name =obj
        else:
            obj_name =obj.name
        self.del_queue.append(obj_name)

    def post_processes(self):
        if self.del_queue:
            for i in self.del_queue:
                del self.objects[i]
                self.objects_order.pop(self.objects_order.index(i))
            self.del_queue = []


    def add_raw(self, func, name, *args):
        self.raw_draw[name] = [func, args]

    def add_blit(self, img, pos):
        self.blit_arr.append((img, pos))
    def update_objects(self):

        if self.pressed_obj:
            pomop = self.pressed_obj.convert_to_local(self.mouse_pos)
            self.pressed_obj.pressed(mouse_pos=pomop, btn_id=self.mouse_state[0])
            self.pressed_obj = None
        if self.released_obj:
            pomop = self.released_obj.convert_to_local(self.mouse_pos)
            self.released_obj.release(mouse_pos=pomop, btn_id=self.mouse_state[0])
            self.released_obj = None
        if self.dragged_obj:
            curr_obj_mouse_pos = self.dragged_obj.convert_to_local(self.mouse_pos)
            if (self.mouse_delta[0] or self.mouse_delta[1]) and self.mouse_state[1]:
                self.dragged_obj.dragged(mouse_pos=curr_obj_mouse_pos,
                                      btn_id=self.mouse_state[0],
                                      mouse_delta=self.mouse_delta)
        if self.curr_obj:
            if (self.mouse_delta[0] or self.mouse_delta[1]) and self.mouse_state[1] and not self.dragged_obj:
                self.dragged_obj = self.curr_obj
            else:
                curr_obj_mouse_pos = self.curr_obj.convert_to_local(self.mouse_pos)
                self.curr_obj.hover(mouse_pos=curr_obj_mouse_pos)

        for obj in self.objects_order:
            self.objects[obj].update()

    def draw_objects(self):
        self.screen.fill(self.bg_color)
        for img, pos in self.blit_arr:
            self.screen.blit(img, pos)
        self.blit_arr = []
        for obj_name in self.objects_order:
            if self.objects[obj_name].visible:
                self.objects[obj_name].draw()
        for func, args in self.raw_draw.values():
            func(self.screen, *args)
        self.raw_draw = dict()

    def sprite(self, sprite_obj):
        while not self.initialized:
            pass
        self.add_object(sprite_obj)
        return self.objects[sprite_obj.name]



import math
import numpy as np


class Projection:
    def __init__(self, render):
        NEAR = render.camera.near_plane
        FAR = render.camera.far_plane
        RIGHT = math.tan(render.camera.h_fov / 2)
        LEFT = -RIGHT
        TOP = math.tan(render.camera.v_fov / 2)
        BOTTOM = -TOP

        m00 = 2 / (RIGHT - LEFT)
        m11 = 2 / (TOP - BOTTOM)
        m22 = (FAR + NEAR) / (FAR - NEAR)
        m32 = -2 * NEAR * FAR / (FAR - NEAR)
        self.projection_matrix = np.array([
            [m00, 0, 0, 0],
            [0, m11, 0, 0],
            [0, 0, m22, 1],
            [0, 0, m32, 0]
        ])

        HW, HH = render.h_width, render.h_height
        self.to_screen_matrix = np.array([
            [HW, 0, 0, 0],
            [0, -HH, 0, 0],
            [0, 0, 1, 0],
            [HW, HH, 0, 1]
        ])


from .Hollow3D import Hollow3D
import numpy as np


class Axis3D(Hollow3D):
    def __init__(self, render):
        super().__init__(render, [(0, 0, 0, 1), (1, 0, 0, 1), (0, 1, 0, 1), (0, 0, -1, 1)], [(0, 1), (0, 2), (0, 3)])
        self.color_edges = np.array([[255, 0, 0], [0, 255, 0], [0, 0, 255]]).astype(np.uint32)
        self.draw_vertices = False
        self.label = 'XYZ'
